from pipa.common.config import CONFIG_DIR
from pipa.common.cmd import run_command
import shutil
import os


def write_string_to_file(string: str, filename: str, append: bool = False):
    file_path = os.path.join(CONFIG_DIR, filename)
    with open(file_path, "a" if append else "w") as file:
        result = file.write(string)
    return result


def copy_file(source_file: str, destination_folder: str = CONFIG_DIR):
    return shutil.copy(source_file, destination_folder)


def run_command_and_write_to_file(command: str, filename: str, append: bool = False):
    result = run_command(command, log=True)
    write_string_to_file(result, filename, append)
    return result



import fire
from pipa.service.gengerate.all import quest_summary as generate_sh


class PipaCLI:
    def generate(self):
        generate_sh()


def main():
    fire.Fire(PipaCLI)


if __name__ == "__main__":
    main()



import questionary

from pipa.service.gengerate.run_by_pipa import main as run_by_pipa
from pipa.service.gengerate.run_by_user import main as run_by_user


def quest_summary():
    how_to_run = questionary.select(
        "Please select the way of workload you want to run.",
        choices=[
            "Build a script that collects performance data and start the workload by perf.",
            "Build a script that collects global performance data.",
        ],
    ).ask()
    if (
        how_to_run
        == "Build a script that collects performance data and start the workload by perf."
    ):
        run_by_pipa()
    elif how_to_run == "Build a script that collects global performance data.":
        run_by_user()



import pytest
from pipa.parser.perf_script import parse_one_line


def test_parse_one_line():
    lines: list = [
        "perf 1585183 [000] 4492421.247039:          1        cpu_core/instructions/:  ffffffff9030a814 perf_ctx_enable+0x24 ([kernel.kallsyms])",
        "perf 1585183 [000] 4492421.247039: 1 cpu_core/instructions/:  ffffffff9030a814 perf_ctx_enable+0x24 ([kernel.kallsyms])",
        " :-1      -1 [001] 4492431.253176:      28168       cpu_core/branch-misses/:  ffffffff90172282 __update_load_avg_se+0xa2 ([kernel.kallsyms])",
        "perf 1097778 [000] 45703.470045:      21326          ref-cycles:      558d5f421ffb  evlist_cpu_iterator__next+0x2b (/usr/lib/linux-tools-5.15.0-89/perf)",
        "perf 1585183 [000] 4492421.247039:    21326          cpu_core/instructions/:  ffffffff9030a814 perf_ctx_enable+0x24 (/kernel/kallsyms)",
        "node 3174706 [028] 1878680.878876:   19098947 branch-instructions:           1e93f03 std::vector<v8::internal::compiler::StateValueDescriptor, v8::internal::ZoneAllocator<v8::internal::compiler::StateValueDescriptor> >::_M_fill_insert+0x113 (/mnt/hdd/users/xyjiang/.vscode-server/bin/0ee08df0cf4527e40edc9aa28f4b5bd38bbff2b2/node)",
    ]

    expected_output_list: list = [
        [
            "perf",
            1585183,
            0,
            "4492421.247039",
            1,
            "cpu_core/instructions/",
            "ffffffff9030a814",
            "perf_ctx_enable+0x24",
            "[kernel.kallsyms]",
        ],
        [
            "perf",
            1585183,
            0,
            "4492421.247039",
            1,
            "cpu_core/instructions/",
            "ffffffff9030a814",
            "perf_ctx_enable+0x24",
            "[kernel.kallsyms]",
        ],
        [
            ":-1",
            -1,
            1,
            "4492431.253176",
            28168,
            "cpu_core/branch-misses/",
            "ffffffff90172282",
            "__update_load_avg_se+0xa2",
            "[kernel.kallsyms]",
        ],
        [
            "perf",
            1097778,
            0,
            "45703.470045",
            21326,
            "ref-cycles",
            "558d5f421ffb",
            "evlist_cpu_iterator__next+0x2b",
            "/usr/lib/linux-tools-5.15.0-89/perf",
        ],
        [
            "perf",
            1585183,
            0,
            "4492421.247039",
            21326,
            "cpu_core/instructions/",
            "ffffffff9030a814",
            "perf_ctx_enable+0x24",
            "/kernel/kallsyms",
        ],
        [
            "node",
            3174706,
            28,
            "1878680.878876",
            19098947,
            "branch-instructions",
            "1e93f03",
            "std::vector<v8::internal::compiler::StateValueDescriptor, v8::internal::ZoneAllocator<v8::internal::compiler::StateValueDescriptor> >::_M_fill_insert+0x113",
            "/mnt/hdd/users/xyjiang/.vscode-server/bin/0ee08df0cf4527e40edc9aa28f4b5bd38bbff2b2/node",
        ],
    ]
    for i, line in enumerate(lines):
        assert parse_one_line(line) == expected_output_list[i]


if __name__ == "__main__":
    pytest.main([__file__])



import questionary
from rich import print
from pipa.service.gengerate.common import quest_basic, CORES_ALL, write_title


def quest():
    """
    This function prompts the user with a series of questions to gather information for running a workload.

    Returns:
        A tuple containing the following information:
        - workspace: The workspace path
        - freq_record: Frequency record
        - events_record: Events record
        - freq_stat: Frequency statistics
        - events_stat: Events statistics
        - annotete: Whether to use perf-annotate (True or False)
        - command: The command of the workload
    """
    workspace, freq_record, events_record, freq_stat, events_stat, annotete = (
        quest_basic()
    )

    use_taskset = questionary.select(
        "Whether to use taskset?\n", choices=["Yes", "No"]
    ).ask()

    if use_taskset == "Yes":
        cores_input = questionary.text(
            "Which cores do you want to use? (Default: 120-140)\n", "120-140"
        ).ask()

        if cores_input.isdigit():
            core_list = cores_input.strip()
        elif cores_input.split("-").__len__() != 2:
            print("Please input a valid range, split by '-'.")
            exit(1)
        else:
            left, right = cores_input.split("-")

            left, right = left.strip(), right.strip()
            if not left.isdigit() or not right.isdigit():
                print("Please input a valid range, non-digit char detected.")
                exit(1)
            left, right = int(left), int(right)
            if left < CORES_ALL[0] or right > CORES_ALL[-1] or left > right:
                print("Please input a valid range.")
                exit(1)
            core_list = ",".join([str(i) for i in list(range(left, right + 1))])

    command = questionary.text("What's the command of workload?\n").ask()

    if command == "":
        print("Please input a command to run workload.")
        exit(1)

    if use_taskset == "Yes":
        command = f"/usr/bin/taskset -c {core_list} {command}"
    return (
        workspace,
        freq_record,
        events_record,
        freq_stat,
        events_stat,
        annotete,
        command,
    )


def generate(
    workspace, freq_record, events_record, freq_stat, events_stat, annotete, command
):
    """
    Generate a shell script for collecting performance data which start workload by perf.

    Args:
        workspace (str): The path to the workspace directory.
        freq_record (str): The frequency for recording events.
        events_record (str): The events to be recorded.
        freq_stat (str): The frequency for collecting statistics.
        events_stat (str): The events to be collected for statistics.
        annotete (bool): Whether to annotate the performance data.
        command (str): The command to be executed.

    Returns:
        None
    """
    with open(workspace + "/pipa-run.sh", "w") as f:
        write_title(f)

        f.write("WORKSPACE=" + workspace + "\n")
        f.write("mkdir -p $WORKSPACE\n\n")

        f.write(
            f"perf record -e '{events_record}' -a -F"
            + f" {freq_record} -o $WORKSPACE/perf.data {command}\n"
        )
        f.write("perf script -i $WORKSPACE/perf.data > $WORKSPACE/perf.script\n")
        f.write("perf report -i $WORKSPACE/perf.data > $WORKSPACE/perf.report\n\n")

        f.write("sar -o $WORKSPACE/sar.dat 1 >/dev/null 2>&1 &\n")
        f.write("sar_pid=$!\n")
        f.write(
            f"perf stat -e {events_stat} -C {CORES_ALL[0]}-{CORES_ALL[-1]} -A -x , -I {freq_stat} -o $WORKSPACE/perf-stat.csv {command}\n"
        )
        f.write("kill -9 $sar_pid\n")
        f.write("sar -A -f $WORKSPACE/sar.dat >$WORKSPACE/sar.txt\n\n")

        if annotete:
            f.write(
                "perf annotate -i $WORKSPACE/perf.data > $WORKSPACE/perf.annotate\n\n"
            )

        f.write("echo 'Performance data collected successfully.'\n")

        print("Shell script generated successfully.")
        print("Please check the script in " + workspace + "/pipa-run.sh")


def main():
    generate(*quest())


if __name__ == "__main__":
    main()



import pandas as pd
import re
from pipa.common.logger import logger
import multiprocessing
from pipa.export_config.cpu_config import NB_PHYSICAL_CORES


def parse_one_line(line):
    """
    Parses a single line of input and returns a list of parsed values.

    Args:
        line (str): The input line to be parsed.

    Returns:
        list: A list containing the parsed values from the input line.
            The list contains the following elements in order:
            - command (str): The command name.
            - pid (int): The process ID.
            - tid (int): The thread ID.
            - cpu (int): The CPU number.
            - time (str): The time value.
            - value (int): The event value.
            - event (str): The event name.
            - addr (str): The address value.
            - symbol (str): The symbol name.
            - caller (str): The caller value.
    """
    try:
        try:
            pattern = r"(\S+|\:-\d+)\s+(\d+|-\d+)\s+\[(\d+)]\s+(\d+\.\d+):\s+(\d+)\s+(\S+):\s+(\S+)\s+(.*?)\s+\((\S+)\)"

            command, pid, cpu, time, value, event, addr, symbol, caller = re.match(
                pattern, line.strip()
            ).groups()
        except:
            pattern = r"(\d+|-\d+)\s+\[(\d+)]\s+(\d+\.\d+):\s+(\d+)\s+(\S+):\s+(\S+)\s+(.*?)\s+\((\S+)\)"
            (
                pid,
                cpu,
                time,
                value,
                event,
                addr,
                symbol,
                caller,
            ) = re.match(pattern, line[15:].strip()).groups()

            command = line[:15].strip()

    except Exception as e:
        logger.warning("parse failed for line: " + line + "\n with error: " + str(e))
        return None

    return [
        command,
        int(pid),
        int(cpu),
        time,
        int(value),
        event,
        addr,
        symbol,
        caller,
    ]


def parse_perf_script_file(parsed_script_path, processes_num=NB_PHYSICAL_CORES):
    """
    Parses a perf script file and returns the data as a pandas DataFrame.

    Args:
        parsed_script_path (str): The path to the perf script file.

    Returns:
        pandas.DataFrame: The parsed data as a DataFrame.
    """

    # Open the perf script file in read mode
    with open(parsed_script_path, "r") as file:
        # Read all lines from the file and remove leading/trailing whitespaces
        content = [l.strip() for l in file.readlines()]

    # Ensure that the content is not None
    if content is None:
        logger.info("content is None")
        return None

    # Define the separator used to split the header and table sections
    SEPARATOR = "# ========"

    # Check if the first line is the separator, and remove it if present
    if content[0] == SEPARATOR:
        content.pop(0)

    if SEPARATOR in content:
        # Find the index of the separator in the content list
        sep_idx = content.index(SEPARATOR)

        # Split the content into header and table sections
        header, table = content[:sep_idx], content[sep_idx + 2 :]
    else:
        table = content

    pool = multiprocessing.Pool(processes=processes_num)
    data = pool.map(parse_one_line, table)
    pool.close()
    pool.join()

    data = [d for d in data if d is not None]

    # Parse each line in the table section and create a DataFrame
    return pd.DataFrame(
        data,
        columns=[
            "command",
            "pid",
            "cpu",
            "time",
            "value",
            "event",
            "addr",
            "symbol",
            "caller",
        ],
    )


if __name__ == "__main__":
    parse_perf_script_file(input("input perf script text data path:"))



import questionary
from pipa.export_config.cpu_config import get_cpu_cores
from rich import print
from io import TextIOWrapper
import os


def ask_number(question: str, default: int) -> int:
    """
    Asks the user to input a number based on the given question and default value.

    Args:
        question (str): The question to ask the user.
        default (int): The default value to return if the user doesn't input anything.

    Returns:
        int: The number inputted by the user or the default value.
    """
    result = questionary.text(question, str(default)).ask().strip()

    if result == "":
        return default
    elif result.isdigit():
        return int(result)
    else:
        print("Please input a number.")
        exit(1)


CORES_ALL = get_cpu_cores()


def quest_basic():
    workspace = questionary.text(
        "Where do you want to store your data? (Default: ./)\n", "./"
    ).ask()

    if workspace == "":
        workspace = "./"

    if not os.path.exists(workspace):
        os.makedirs(workspace)

    freq_record = ask_number("What's the frequency of perf-record? (Default: 99)\n", 99)
    events_record = questionary.text(
        "What's the event of perf-record? (Default: {cycles,instructions}:S)\n",
        "{cycles,instructions}:S",
    ).ask()

    freq_stat = ask_number(
        "What's count deltas of perf-stat? (Default: 1000 milliseconds)\n", 1000
    )
    events_stat = questionary.text(
        "What's the event of perf-stat?\n (Default: cycles,instructions,LLC-loads,LLC-load-misses,LLC-stores,LLC-store-misses,branch-load-misses,branch-loads)\n",
        "cycles,instructions,LLC-loads,LLC-load-misses,LLC-stores,LLC-store-misses,branch-load-misses,branch-loads",
    ).ask()

    annotete = questionary.select(
        "Whether to use perf-annotate?\n", choices=["Yes", "No"]
    ).ask()
    annotete = True if annotete == "Yes" else False

    return workspace, freq_record, events_record, freq_stat, events_stat, annotete


def write_title(file: TextIOWrapper):
    file.write(
        """
#!/bin/bash
# The script generated by PIPA-TREE is used to collect performance data. 
# Please check whether it meets expectations before running.
# ZJU SPAIL(https://github.com/ZJU-SPAIL) reserves all rights.

# Check if sar and perf are available
if ! command -v sar &> /dev/null; then
echo "sar command not found. Please install sar."
exit 1
fi

if ! command -v perf &> /dev/null; then
echo "perf command not found. Please install perf."
exit 1
fi\n\n"""
    )



from pipa.common.cmd import run_command
from pipa.export_config.utils import write_string_to_file, copy_file
import psutil

def get_lscpu_info():
    result = (
        run_command("lscpu", log=True)
        + "\n\n\n"
        + run_command("lscpu -a --extended", log=True)
    )
    return write_string_to_file(result, "cpu_info.txt")


def get_cpuinfo():
    return copy_file("/proc/cpuinfo")


def get_all_cpu_config():
    get_lscpu_info()
    get_cpuinfo()


def get_cpu_cores():
    cpu_list = [
        l
        for l in run_command("lscpu -p=cpu", log=False).split("\n")
        if not l.startswith("#")
    ]
    return [int(x) for x in cpu_list]

NB_PHYSICAL_CORES = psutil.cpu_count(logical=False)

if __name__ == "__main__":
    get_all_cpu_config()



from pipa.parser.perf_script import parse_perf_script_file
from pipa.parser.perf_stat import parse_perf_stat_file
from pipa.parser.sar import parse_sar_bin
from pipa.service.collect import collect_perf_record, collect_perf_stat_with_sar


def run_and_collect_all(workload_cmd: str):
    perf_stat_dump_path, sar_dump_path = collect_perf_stat_with_sar(
        workload_cmd=workload_cmd
    )
    sar_df_list = parse_sar_bin(sar_dump_path)
    perf_stat_df = parse_perf_stat_file(perf_stat_dump_path)

    report_dump_path, script_dump_path = collect_perf_record(workload_cmd=workload_cmd)
    perf_script_df = parse_perf_script_file(script_dump_path)
    return sar_df_list, perf_stat_df, perf_script_df


if __name__ == "__main__":
    sar_df_list, perf_stat_df, perf_script_df = run_and_collect_all(
        "perf bench futex hash"
    )
    print(sar_df_list)
    print(perf_stat_df)
    print(perf_script_df)



from pipa.parser.sar import split_sar_block
import pytest


def test_split_sar_block():
    # Test splitting a block with multiple lines
    lines = ["line 1\n", "line 2\n", "\n", "line 3\n", "\n", "line 4\n"]
    expected_output = [["line 1", "line 2"], ["line 3"], ["line 4"]]
    assert split_sar_block(lines) == expected_output

    # Test splitting a block with multiple '\n'
    lines = ["line 1\n", "line 2\n", "\n", "\n", "line 3\n", "\n", "line 4\n"]
    expected_output = [["line 1", "line 2"], ["line 3"], ["line 4"]]
    assert split_sar_block(lines) == expected_output

    # Test splitting a block with a single line
    lines = ["line 1\n", "\n"]
    expected_output = [["line 1"]]
    assert split_sar_block(lines) == expected_output

    # Test splitting an empty block
    lines = []
    expected_output = [[]]
    assert split_sar_block(lines) == expected_output

    # Test splitting a block with no empty lines
    lines = ["line 1\n", "line 2\n", "line 3\n"]
    expected_output = [["line 1", "line 2", "line 3"]]
    assert split_sar_block(lines) == expected_output


if __name__ == "__main__":
    pytest.main([__file__])



import platform
import yaml
from pipa.common.logger import logger
from pipa.common.config import CONFIG_DIR


def get_platform_info():
    info = {
        "architecture": str(platform.architecture()),
        "machine": platform.machine(),
        "node": platform.node(),
        "platform": platform.platform(aliased=False, terse=False),
        "processor": platform.processor(),
        "python_build": str(platform.python_build()),
        "python_compiler": platform.python_compiler(),
        "python_branch": platform.python_branch(),
        "python_implementation": platform.python_implementation(),
        "python_revision": platform.python_revision(),
        "python_version": platform.python_version(),
        "release": platform.release(),
        "system": platform.system(),
        "version": platform.version(),
    }
    logger.info("Platform information: %s", info)
    yaml.dump(info, open(f"{CONFIG_DIR}/python_platform_info.yaml", "w"))
    return info

if __name__ == "__main__":
    get_platform_info()


from pipa.export_config.cpu_config import get_all_cpu_config
from pipa.export_config.py_config import get_platform_info
from pipa.export_config.sys_config import get_all_sys_config


def get_all_config():
    """
    Retrieves all configuration information.

    This function calls several other functions to gather platform information,
    system configuration, and CPU configuration.

    Returns:
        None
    """
    get_platform_info()
    get_all_sys_config()
    get_all_cpu_config()


if __name__ == "__main__":
    get_all_config()



import questionary
from rich import print
from pipa.service.gengerate.common import quest_basic, CORES_ALL, write_title


def quest():
    workspace, freq_record, events_record, freq_stat, events_stat, annotete = (
        quest_basic()
    )

    set_record = questionary.select(
        "Whether to set the duration of the perf-record run?\n",
        choices=["Yes", "No, I'll control it by myself. (Exit by Ctrl+C)"],
    ).ask()
    record_time, stat_time = None, None
    if set_record == "Yes":
        record_time = questionary.text(
            "How long do you want to run perf-record? (Default: 10s)\n", "10"
        ).ask()

    set_stat = questionary.select(
        "Whether to set the duration of the perf-stat run?\n",
        choices=["Yes", "No, I'll control it by myself. (Exit by Ctrl+C)"],
    ).ask()
    if set_stat == "Yes":
        stat_time = questionary.text(
            "How long do you want to run perf-stat? (Default: 10s)\n", "10"
        ).ask()
    return (
        workspace,
        freq_record,
        events_record,
        freq_stat,
        events_stat,
        annotete,
        record_time,
        stat_time,
    )


def generate(
    workspace,
    freq_record,
    events_record,
    freq_stat,
    events_stat,
    annotete,
    record_time,
    stat_time,
):
    with open(workspace + "/pipa-collect.sh", "w") as f:
        write_title(f)

        f.write("WORKSPACE=" + workspace + "\n")
        f.write("mkdir -p $WORKSPACE\n\n")

        f.write(
            f"perf record -e '{events_record}' -a -F"
            + f" {freq_record} -o $WORKSPACE/perf.data"
            + f" -- sleep {record_time}\n"
            if record_time
            else "\n"
        )
        f.write("perf script -i $WORKSPACE/perf.data > $WORKSPACE/perf.script\n")
        f.write("perf report -i $WORKSPACE/perf.data > $WORKSPACE/perf.report\n\n")

        f.write("sar -o $WORKSPACE/sar.dat 1 >/dev/null 2>&1 &\n")
        f.write("sar_pid=$!\n")
        f.write(
            f"perf stat -e {events_stat} -C {CORES_ALL[0]}-{CORES_ALL[-1]} -A -x , -I {freq_stat} -o $WORKSPACE/perf-stat.csv"
            + f" sleep {stat_time}\n"
            if stat_time
            else "\n"
        )
        f.write("kill -9 $sar_pid\n")
        f.write("sar -A -f $WORKSPACE/sar.dat >$WORKSPACE/sar.txt\n\n")

        if annotete:
            f.write(
                "perf annotate -i $WORKSPACE/perf.data > $WORKSPACE/perf.annotate\n\n"
            )

        f.write("echo 'Performance data collected successfully.'\n")

        print("Shell script generated successfully.")
        print("Please check the script in " + workspace + "/pipa-collect.sh")


def main():
    generate(*quest())


if __name__ == "__main__":
    main()



from pipa.export_config.utils import run_command_and_write_to_file, copy_file


def get_dmidecode():
    return run_command_and_write_to_file("sudo dmidecode", "dmidecode.txt")


def get_lspci():
    return run_command_and_write_to_file("lspci", "lspci.txt")


def get_lsusb():
    return run_command_and_write_to_file("lsusb", "lsusb.txt")


def get_disk_usage():
    return run_command_and_write_to_file("df -h", "disk_usage.txt")


def get_interrupt_info():
    return copy_file("/proc/interrupts")


def get_meminfo():
    return copy_file("/proc/meminfo")


def get_version():
    return copy_file("/proc/version")


def get_all_sys_config():
    get_dmidecode()
    get_lspci()
    get_lsusb()
    get_disk_usage()
    get_interrupt_info()
    get_meminfo()
    get_version()


if __name__ == "__main__":
    get_all_sys_config()



import pandas as pd
from pandarallel import pandarallel, NB_PHYSICAL_CORES
import seaborn as sns


class PerfStatData:
    def __init__(self, perf_stat_csv_path: str):
        self.data = parse_perf_stat_file(perf_stat_csv_path)

    def get_CPI(self):
        """
        Returns the CPI (Cycles Per Instruction) data.

        Returns:
            pd.DataFrame: Dataframe containing the CPI data.
        """
        return (
            self.data[self.data["metric_type"] == "cycles"]
            .merge(
                self.data[self.data["metric_type"] == "instructions"],
                on=["timestamp", "cpu_id"],
                suffixes=("_cycles", "_instructions"),
            )
            .assign(CPI=lambda x: x["value_cycles"] / x["value_instructions"])
            .drop(columns=["metric_type_cycles", "metric_type_instructions"])
        )

    def get_CPI_time(self, threads: list = None):
        """
        Returns the CPI (Cycles Per Instruction) over time for the specified threads.

        Args:
            threads (list, optional): A list of thread IDs. If None, returns the average CPI over time for all threads.

        Returns:
            pandas.DataFrame: A DataFrame containing the timestamp and CPI values over time.

        """
        if threads is None:
            return self.get_CPI()[["timestamp", "CPI"]].groupby("timestamp").mean()
        df = self.get_CPI()
        return df[df["cpu_id"].isin([int(t) for t in threads])]

    def get_CPI_overall(self, data_type="thread"):
        """
        Calculate the overall CPI (Cycles Per Instruction) based on the given data type.

        Parameters:
        - data_type (str): The type of data to calculate CPI for. Can be "thread" or "system".

        Returns:
        - If data_type is "thread", returns a DataFrame with CPI values per thread.
        - If data_type is "system", returns the overall CPI value for the system.

        Raises:
        - ValueError: If an invalid data type is provided.
        """

        df = self.get_CPI()
        match data_type:
            case "thread":
                data_per_thread = (
                    df[["cpu_id", "value_cycles", "value_instructions"]]
                    .groupby("cpu_id")
                    .sum()
                )
                data_per_thread["CPI"] = (
                    data_per_thread["value_cycles"]
                    / data_per_thread["value_instructions"]
                )
                return data_per_thread
            case "system":
                total_instructions = df["value_instructions"].sum()
                total_cycles = df["value_cycles"].sum()
                return total_cycles / total_instructions
            case _:
                raise ValueError("Invalid data type")

    def plot_CPI_time_by_thread(self, threads: list):
        """
        Plots CPI over time for the specified threads.

        Args:
            threads (list): A list of thread IDs.

        Returns:
            None
        """
        sns.set_theme(style="darkgrid", rc={"figure.figsize": (15, 8)})
        if len(threads) > 1:
            p = sns.lineplot(
                data=self.get_CPI_time(threads), x="timestamp", hue="cpu_id", y="CPI"
            )
        else:
            p = sns.lineplot(data=self.get_CPI_time(threads), x="timestamp", y="CPI")
        p.set_title("CPI over Time, Thread " + ",".join([str(t) for t in threads]))
        p.set_xlabel("Time(s)")
        p.set_ylabel("CPI")
        return p

    def plot_CPI_time_system(self):
        """
        Plots CPI (Cycles Per Instruction) over time for the system.

        This method generates a line plot showing the CPI values over time for the system.
        It uses the data returned by the `get_CPI_time` method and saves the plot as an image.

        Returns:
            None
        """
        sns.set_theme(style="darkgrid", rc={"figure.figsize": (15, 8)})
        p = sns.lineplot(data=self.get_CPI_time(), x="timestamp", y="CPI")
        p.set_title("CPI over Time, System")
        p.set_xlabel("Time(s)")
        p.set_ylabel("CPI")
        return p


def parse_perf_stat_file(stat_output_path: str):
    """
    Parse the perf stat output file and return a pandas DataFrame.

    Args:
        stat_output_path (str): The path to the perf stat output file.

    Returns:
        pandas.DataFrame: The parsed data as a DataFrame.

    The fields are in this order:
    •   optional usec time stamp in fractions of second (with -I xxx)
    •   optional CPU, core, or socket identifier
    •   optional number of logical CPUs aggregated
    •   counter value
    •   unit of the counter value or empty
    •   event name
    •   run time of counter
    •   percentage of measurement time the counter was running
    •   optional metric value
    •   optional unit of metric
    """
    pandarallel.initialize(min(12, NB_PHYSICAL_CORES))
    df = pd.read_csv(
        stat_output_path,
        skiprows=1,
        names=[
            "timestamp",
            "cpu_id",
            "value",
            "unit",
            "metric_type",
            "run_time(ns)",
            "run_percentage",
            "opt_value",
            "opt_unit_metric",
        ],
    ).astype(
        {
            "timestamp": "float64",
            "cpu_id": str,
            "value": "int64",
            "unit": str,
            "metric_type": str,
            "run_time(ns)": "int64",
            "run_percentage": "float64",
            "opt_value": "float64",
            "opt_unit_metric": str,
        }
    )
    df["cpu_id"] = df["cpu_id"].str.removeprefix("CPU").astype(int)
    return df



import os
from pipa.common.cmd import run_command
from pipa.common.config import DUMP_DIR
from pipa.common.utils import get_timestamp
from pipa.service.init import create_directories

create_directories()

def collect_perf_record(
    workload_cmd="perf bench futex hash",
    frequency=97,
    script_path=None,
    events="instructions,ref-cycles,cpu-cycles,branch-instructions,branch-misses",
):
    """
    Collects performance data using the 'perf' command.

    Args:
        workload_cmd (str, optional): The command to execute for workload. Defaults to "perf bench futex hash".
        frequency (int, optional): The frequency at which to sample events. Defaults to 97.
        output_file (str, optional): The path to the output file. If not provided, a default file name will be generated. Defaults to None.
        events (str, optional): The events to record. Defaults to "instructions,ref-cycles,cpu-cycles,branch-instructions,branch-misses".
    """
    now = get_timestamp()
    if not script_path:
        script_path = os.path.join(
            DUMP_DIR, f"workload_{now}.log"
        )  # TODO rename to perf
    # Construct the command to record the CPU events
    run_command(
        f"perf record -e {events} -a -F {frequency} {workload_cmd} > {script_path}"
    )
    # Generate a performance report
    report_path = os.path.join(DUMP_DIR, f"perf_{now}.report")
    run_command(f"perf report --header > {report_path}")
    # Generate output of the perf script
    script_path = os.path.join(
        DUMP_DIR, f"perf_{now}.script"
    )
    run_command(f"perf script --header > {script_path}")
    os.rename("perf.data", f"{DUMP_DIR}/perf_{now}.data")
    return report_path, script_path


def collect_perf_stat(
    workload_cmd="perf bench futex hash",
    sample_rate=997,
    perf_stat_output_path=None,
    events="instructions,ref-cycles,cpu-cycles,branch-instructions,branch-misses",
    core_range="0-19",
):
    """
    Collects performance statistics using the 'perf stat' command.

    Args:
        workload_cmd (str, optional): The command to run for workload. Defaults to "perf bench futex hash".
        sample_rate (int, optional): The sample rate for collecting statistics. Defaults to 997.
        perf_stat_output_path (str, optional): The output path for the perf stat results. If not provided, a default path will be used.
        events (str, optional): The events to measure. Defaults to "instructions,ref-cycles,cpu-cycles,branch-instructions,branch-misses".
        core_range (str, optional): The range of CPU cores to use. Defaults to "0-19".

    Returns:
        str: The path to the perf stat output file.
    """
    now = get_timestamp()
    perf_stat_output_path = perf_stat_output_path or os.path.join(
        DUMP_DIR, f"perf_stat_{now}.csv"
    )
    command = f"perf stat -e {events} -C {core_range} -A -x , -I {sample_rate} -o {perf_stat_output_path} {workload_cmd}"
    run_command(command)
    return perf_stat_output_path


def collect_perf_stat_with_sar(
    workload_cmd="perf bench futex hash",
    sample_rate=1000,
    perf_stat_output_path=None,
    sar_output_path=None,
    events="instructions,ref-cycles,cpu-cycles,branch-instructions,branch-misses",
    core_range="0-19",
):
    """
    Collects performance statistics using perf stat and sar commands.

    Args:
        workload_cmd (str, optional): Command to run for workload. Defaults to "perf bench futex hash".
        sample_rate (int, optional): Sampling rate in milliseconds. Defaults to 1000.
        perf_stat_output_path (str, optional): Path to save perf stat output. Defaults to None.
        sar_output_path (str, optional): Path to save sar output. Defaults to None.
        events (str, optional): Comma-separated list of events to monitor. Defaults to "instructions,ref-cycles,cpu-cycles,branch-instructions,branch-misses".
        core_range (str, optional): Range of CPU cores to monitor. Defaults to "0-19".

    Returns:
        Tuple[str, str]: Path to perf stat output file and path to sar output file.
    """
    sar_output_path = sar_output_path or f"data/dump/sar_output_{get_timestamp()}.dat"
    run_command(f"sar -o {sar_output_path} 1 >/dev/null 2>&1 &")
    perf_stat_output_path = collect_perf_stat(
        workload_cmd, sample_rate, perf_stat_output_path, events, core_range
    )
    run_command("killall sar")
    return perf_stat_output_path, sar_output_path


if __name__ == "__main__":
    collect_perf_stat_with_sar()
    collect_perf_record()
